"""
Module for signal visualization.

Visualize signals in real time with spectral filtering, common average
filtering options and real-time FFT.
"""

from .stream_viewer import StreamViewer  # noqa: F401
